## -----------------------------------
## This is a global library for pyxcom
## -----------------------------------
## Copyright (c) 2004-2005 Alessandro Pisa <darkmoon@altervista.org>
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions
## are met:
##
## 1. Redistributions of source code must retain the above copyright
##    notice, this list of conditions and the following disclaimer.
## 2. Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##
## THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
## IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
## OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
## IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
## INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
## NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
## THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os, sys, getopt

# URL to xcom package at nist website
XCOM_URL='http://physics.nist.gov/PhysRefData/Xcom/XCOM.tar.gz'


# Name and meaning of the colums
COLUMNS_LEGEND=("Photon Energy (MeV)",
                'Coherent Scattering (cm2/g)',
                'Incoherent Scattering (cm2/g)',
                'Photoelectric Absorption (cm2/g)',
                'Pair Production In Nuclear Field (cm2/g)',
                'Pair Production In Electron Field (cm2/g)',
                'Total Attenuation With Coherent Scattering (cm2/g)',
                'Total Attenuation Without Coherent Scattering (cm2/g)')

# Name of the configuration file. At the moment there is no chance to use
# something different from "$HOME/.pyxcomrc"
CFG_FILE = os.path.join(os.environ['HOME'], '.pyxcomrc')

# Default content for the configuration file: it supposes that the program xcom
# is installed under /opt
CFG_DEFAULT = """# pyxcom configuration file
# It is rather simple: just set the the XCOMDIR and XCOMBIN variables
# to the correct values!
# For example if you have XCOM installed in the directory /opt/xcom and
# the executable is called 'XcOm' set these values to:
# XCOMDIR=/opt/xcom
# XCOMBIN=/opt/xcom/XcOm
#
"""

# Definition for this variables
XCOMDIR = ""
XCOMBIN = ""

COLUMNS_MEANING="""\
    1 - Photon Energy (MeV)
    2 - Coherent Scattering (cm2/g)
    3 - Incoherent Scattering (cm2/g)
    4 - Photoelectric Absorption (cm2/g)
    5 - Pair Production In Nuclear Field (cm2/g)
    6 - Pair Production In Electron Field (cm2/g)
    7 - Total Attenuation With Coherent Scattering (cm2/g)
    8 - Total Attenuation Without Coherent Scattering (cm2/g)
"""

## option parsing function
def helpcol():
    """USAGE: helpcol()

    Message printed when --helpcol flag is used.
    After displaying the message the program exits.
    """
    print """COLUMNS MEANING:"""
    print COLUMNS_MEANING
    sys.exit()
        
def getopt_column(optarg):
    """USAGE: getopt_column(optarg)

    Called from the main file while checking from options.
    optarg should be in the form [n1,..,nN].
    Really the parenteses can be substituted with everything
    else because the first and the last character of the optarg
    are trimmed away.
    The commas, instead, are really important and cannot be
    substituted by any other character.
    """
    try:
        # -9 is needed because when edges occur the edge name is specified
        column_list=map(lambda x: int(x)-9, optarg[1:-1].split(','))
    except ValueError:
        ERROR_COLUMN_FLAG="""Syntax error in option --column
        The correct syntax is: --column [colN1,colN2,colN3,...]
        Notice the comma separated values!
        To see what the column number stands run: pyxcom --helpcol
        """.strip()
        sys.exit()
    if False in [-9<x<0 for x in column_list]:
        sys.stderr.write('Columns must vary in the range 1-8!\n')
        sys.exit()
    else:
        return column_list
## End option parsing function

## Configuration check
def isyes(answer):
    """USAGE: isyes(string)

    If the string, converted to lowercase is one of '', 'y' or 'yes'
    returns True, otherwise False
    """
    YES=('', 'y', 'yes')
    return answer.lower() in YES

def ask_xcomdir():
    """USAGE: ask_xcomdir()

    Asks for the directory where xcom is installed until it is a correct directory
    """
    sys.stderr.write("Please enter the full path to the directory where XCOM executable is installed: ")
    XCOMDIR=raw_input()
    while not os.path.isdir(XCOMDIR):
        sys.stderr.write("%s appears not to be a valid directory\n""" % XCOMDIR)
        sys.stderr.write("Please enter the full path to the directory where XCOM executable is installed: ")
        XCOMDIR=raw_input()
    return XCOMDIR

def ask_xcombin(XCOMDIR):
    """USAGE: ask_xcombin(XCOMDIR)

    Asks for the name of the xcom executable until it is a file in the XCOMDIR directory
    """
    sys.stderr.write("""Enter the name of the executable: """)
    XCOMBIN=raw_input()
    XCOMBIN=os.path.join(XCOMDIR,XCOMBIN)
    while not os.path.isfile(XCOMBIN):
        sys.stderr.write("""%s appears not to be a valid file\n""" % XCOMBIN)
        sys.stderr.write("""Enter the name of the Xcom executable: """)
        XCOMBIN=raw_input()
    return XCOMBIN

def update_conf(XCOMDIR, XCOMBIN):
    """USAGE: update_conf(XCOMDIR, XCOMBIN)

    Updates the configuration file $HOME/.pyxcomrc to new values for XCOMDIR & XCOMBIN
    """
    new_conf=[]
    if os.path.isfile(CFG_FILE):
        lines=file(CFG_FILE).readlines()
        for line in lines:
            if len(line.strip()) > 0:
                if line.strip()[0] == '#':new_conf.append(line)
    new_conf.append('XCOMDIR=%s\n' % XCOMDIR)
    new_conf.append('XCOMBIN=%s\n' % XCOMBIN)
    file(CFG_FILE, 'w').writelines(new_conf)

def check_conf():
    """USAGE check_conf()

    Checks if the file CFG_FILE exists and is valid.
    If not provides it"""
    if not os.path.isfile(CFG_FILE):
        sys.stderr.write("""No configuration file found.\n""")
        sys.stderr.write("""Do you need to install xcom (wget needed)? [Y/n]""")
        DOWNLOADXCOM=raw_input()
        if isyes(DOWNLOADXCOM):
            XCOMDIR, XCOMBIN = install_xcom()
        else:
            # Asking for path for an already installed version of xcom
            XCOMDIR=ask_xcomdir()
            XCOMBIN=ask_xcombin(XCOMDIR)
        lines=CFG_DEFAULT.splitlines(True)
        lines.append("XCOMDIR=%s\n" % XCOMDIR)
        lines.append("XCOMBIN=%s\n" % XCOMBIN)
        file(CFG_FILE, 'w').writelines(lines)
    else:
        # Check if XCOMDIR and XCOMBIN are set to correct values and
        # eventually corrects them
        XCOMDIR, XCOMBIN = read_conf()
        
def read_conf():
    """USAGE: read_conf()

    Reads the CFG_FILE and checks if it is correct.
    If it is not corrects it.
    """
    if not os.path.isfile(CFG_FILE): return '',''
    XCOMDIR, XCOMBIN = None, None
    for line in file(CFG_FILE):
        if '=' in line:
            key, value=line.split('=')
            if key.strip() == 'XCOMDIR':XCOMDIR=value[:-1].strip()
            elif key.strip() == 'XCOMBIN': XCOMBIN=value[:-1].strip()
    if XCOMDIR is None: XCOMDIR=ask_xcomdir()
    if XCOMBIN is None: XCOMBIN=ask_xcombin(XCOMDIR)
    update_conf(XCOMDIR, XCOMBIN)
    return XCOMDIR, XCOMBIN
## END Configuration check

## xcom installation
def install_xcom():
    """USAGE: install_xcom()

    Downloads, extracts, compiles xcom to the directory of your choice
    """
    # alias for sys.stderr.write
    say=sys.stderr.write

    # warning on permission
    say("""
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    Root privileges may be need to install xcom in directory where you do
    not have write permission.
    You can:
    1 - install xcom manually;
    2 - install as root or ask root to install the package;
    3 - run pyxcom as root to use this authomatized procedure.
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    """)

    # wgetting xcom
    say('Getting xcom with wget.\n')
    say('Enter a directory where the source will be temporarily stored[/tmp]:')
    TMPDIR=raw_input() or '/tmp'
    while not os.path.isdir(TMPDIR):
        try:
            say('Please create the directory and press enter.')
            say('(Ctrl-C to exit)')
            raw_input()
        except KeyboardInterrupt:
            sys.exit()
    os.system('wget %s -P %s' % (XCOM_URL, TMPDIR))
              
    # Extracting 
    say("Where do you want to install xcom (absolute path) [/opt]? ")
    XCOM_PARENT_DIR=raw_input() or '/opt'
    while not os.path.isdir(XCOM_PARENT_DIR):
        try:
            say('Please create the directory and press enter or Ctrl-C to exit)')
            raw_input()
        except KeyboardInterrupt:
            sys.exit()
    say("Unpacking xcom\n")
    os.system('tar xzf %s/XCOM.tar.gz -C %s' % (TMPDIR, XCOM_PARENT_DIR))

    # remove the sources if needed
    say("""Do you want to remove the downloaded package? [Y/n]""")
    REMOVE=raw_input() or ''
    if isyes(REMOVE):
        os.remove(os.path.join(TMPDIR,'XCOM.tar.gz'))

    # Rename XCOM directory (I like it this way)
    os.chdir(XCOM_PARENT_DIR)
    os.rename('XCOM', 'xcom')
    XCOMDIR=os.path.join(XCOM_PARENT_DIR, 'xcom')
    os.chdir(XCOMDIR)
    # Compile xcom
    os.system('g77 XCOM.f -o xcom')
    # Set xcom command
    XCOMBIN=os.path.join(XCOMDIR, 'xcom')
    return XCOMDIR, XCOMBIN
## END xcom installation        

## cmdline creation
# Here there  functions that decide what to ask to xcom to obtain
# correct data.
# One is for element and the other for compounds
def keV2MeV(keV):
    """USAGE: keV2MeV(keV)

    ...
    """
    return str(float(keV)/1000)

def element_cmdline(keV, flags):
    """USAGE: get_attenuation_element(keV)

    Interacts with XCOM to have values for absortion for a given element
    """
    MeV=map(keV2MeV, keV)
    MeVlen=str(len(MeV))
    cmdline="\n".join(("None",
             '1',
             flags['Z'],
             '3',
             '3',
             '1',
             MeVlen,
             '\n'.join(MeV),
             'N',
             '/dev/stdout',
             '1'
             ,''))
    return cmdline

def compound_cmdline(keV, flags):
    """USAGE: get_attenuation_compound(keV)

    Interacts with XCOM to have values for absortion for a given compound
    """
    MeV=map(lambda x: str(float(x)/1000), keV)
    MeVlen=str(len(MeV))
    cmdline="\n".join(("None",
                       '3',
                       flags['mat'],
                       '3',
                       '1',
                       MeVlen,
                       '\n'.join(MeV),
                       'N',
                       '/dev/stdout',
                       '1',
                       ''))
    return cmdline

def stdelement_cmdline(keV, flags):
    cmdline="\n".join(("None",
                       '1',
                       flags['Z'],
                       '3',
                       '1',
                       '/dev/stdout',
                       '1'
                       ,''))
    return cmdline

def stdcompound_cmdline(keV, flags):
    cmdline="\n".join(("None",
                       '3',
                       flags['mat'],
                       '1',
                       '/dev/stdout',
                       '1'
                       ,''))
    return cmdline
## End cmdline composition

def get_needed(xcom_output, flags):
    """USAGE: get_needed(xcom_output)

    Extracts from xcom_output the value that are specified with --columns flag
    """
    needed=[]
    for line in xcom_output:
        try:
            # The correct lines are the ones containing 8 float at the end
            # Some lines can contain the label of the photo electric edge
            # at the beginning
            line_splitted=map(float, line.split()[-8:])
            line_needed=""
            if len(line_splitted) >=8:
                for i in flags['columns']:
                    line_needed+="%10s" % line_splitted[i]
                needed.append(line_needed)
        except ValueError:
            pass
    return needed

def cmdline2table(cmdline, XCOMBIN, flags):
    """USAGE: cmdline2table(cmdline)

    takes the command line generated by $WHATYOUWANT_cmdline functions
    and uses it to obtain values from xcom. The output of xcom is then
    parsed to extract only the information needed.
    """
    xcom_output=os.popen('echo "%s"|%s ' % (cmdline, XCOMBIN)).readlines()
    # Getting only the total attenuation
    table = get_needed(xcom_output, flags)
    return table

def headers(flags, cmdline):
    """USAGE: comment_lines(flags)

    Comments the lines of the columns from the flag 'columns'
    """
    field1 = "# XCOM COMMANDS:\n"
    combed_cmdline= "\n".join(map(lambda x: '# '+ x, cmdline.split('\n')))+ '\n'
    field2 = '# COLUMNS DESCRIPTION:\n'
    col_str=''
    for i, col in enumerate(flags['columns']):
        col_str+="# %s: %s\n" % (i+1, COLUMNS_LEGEND[col])
    field3 = '#\n# DATA:'
    header = "%s%s%s%s%s" % (field1, combed_cmdline, field2, col_str, field3)
    return header

def text_mode(args, flags):
    """USAGE: text_mode(opts, args, flags)

    uses this program from commandline, output on stdout
    """
    check_conf()
    XCOMDIR, XCOMBIN = read_conf()
    OLDDIR=os.getcwd()
    os.chdir(XCOMDIR)
    if flags['range']:
        Emin, Emax = map(int, flags['range'][1:-1].split(':'))
        args=range(Emin, Emax)
    if flags['standard'] or len(args)==0:
        if flags['mat']:
            cmdline=stdcompound_cmdline(args, flags)
        else:
            cmdline=stdelement_cmdline(args, flags)
    else:
        if not flags['mat']:
            cmdline=element_cmdline(args, flags)
        else:
            cmdline=compound_cmdline(args, flags)
    table=cmdline2table(cmdline, XCOMBIN, flags)
    os.chdir(OLDDIR)
    return headers(flags, cmdline)+"\n"+"\n".join(table)

       
